/*
 * Copyright (C) 2002 Amit Singh <amit_s@acm.org>. All Rights Reserved.
 *
 * Thanks to Brian Mihulka <bmihulka@hulkster.net> for his Emptris
 * application, which I looked at to figure out how to use the Empeg
 * display.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/time.h>

#include "nrinc.h"
#include "emphanoi.h"
#include "ehfont.h"
#include "gravity.h"

static tower_t
towers[3] = {
    { 0, 0, T0xc, -1 },
    { 1, 0, T1xc, -1 },
    { 2, 0, T2xc, -1 },
};

/* Empeg display device global variables */
int fd;
caddr_t dmap;
int **display;

/* Empeg control device global variables */
int ctrl;

/* Hanoi global variables */
long int N;
static int eh_moves = 0;
static int slowdown = 0; /* maximum is MAXSLOWDOWN */
static int movedelay = 0;

int
main(int argc, char **argv)
{
    int i, ret;
    fd_set read_fd_set;
    struct timeval tv;
    char buttonb[4];
    int N_cached;
	
    if (argc != 2) {
        fprintf(stderr, "usage: %s N\n", argv[0]);
        exit(1);
    }

    N = strtol(argv[1], (char **)NULL, 10);
    N_cached = N;

    /* a bit of error checking, LONG_XXX is there in limits.h */
    if (N == LONG_MIN || N == LONG_MAX || N <= 0) {
      fprintf(stderr, "illegal value for number of disks\n");
      exit(2);
    }

    if (N > MAXDISKS) {
        fprintf(stderr, "number of disks exceeded (maximum is %d)\n", MAXDISKS);
        exit(3);
    }

    display = imatrix(0, EMPEG_XMAX, 0, EMPEG_YMAX);
    cleardisplay();
	
    if ((fd = open(DISPLAYDEV, O_RDWR)) < 0) {
        fprintf(stderr, "failed to open %s.\n", DISPLAYDEV);
        exit(4);
    }

    dmap = mmap(0, 2048, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (dmap == (caddr_t)-1) {
        fprintf(stderr, "failed to mmap() from %s.\n", DISPLAYDEV);
        close(fd);
        exit(5);
    }
		
    if ((ctrl = open(CTRLDEV, O_RDONLY | O_NDELAY)) < 0) {
        fprintf(stderr, "failed to open %s.\n", CTRLDEV);
        exit(6);
    }
		
    DISPLAY(fd, 1);

    createtower(0);
    createtower(1);
    createtower(2);

    showlogo();
    showspeed();

    for (i = 0; i < N; i++)
        createdisk(i, 0);

    updatedisplay();
    BLIT(fd);

    while (1) {

        tv.tv_sec = 5;
        tv.tv_usec = 0;

        FD_ZERO(&read_fd_set);
        FD_SET(ctrl, &read_fd_set);

        // if ((ret = select(ctrl + 1, &read_fd_set, NULL, NULL, &tv)) < 0) {
        if ((ret = select(ctrl + 1, &read_fd_set, NULL, NULL, NULL)) < 0) {
            perror("select");
            exit(7);
        }

        if (ret == 0) {
            continue;
        }

        if (!FD_ISSET(ctrl, &read_fd_set)) {
            continue;
        }

        ret = read(ctrl, (char *)buttonb, 4);
        if (ret != 4) {
            continue;
        }

        if ((buttonb[1] != 0) || (buttonb[2] != 0) || (buttonb[3] != 0)) {
            continue;
        }

#ifdef DEBUG
        printf("0x%x 0x%x 0x%x 0x%x\n", buttonb[0], buttonb[1],
                                        buttonb[2], buttonb[3]);
#endif

        switch (buttonb[0]) {
        case VOLUME_UP:
            if (N < MAXDISKS) {
                createdisk(N, 0);
                N++;
                updatedisplay();
                BLIT(fd);
            }
            break;

        case VOLUME_DOWN:
            if (N > 1) {
                N--;
                popdisk(0);
                updatedisplay();
                BLIT(fd);
            }
            break;

        case VOLUME_PRESS:
            goto HANOI;
            break;

        case RESET_PRESS:
            if (N > N_cached) {
                for (; N != N_cached; N--)
                    popdisk(0);
            } else if (N < N_cached) {
                for (; N != N_cached; N++)
                    createdisk(N, 0);
            }
            updatedisplay();
            BLIT(fd);
            break;

        case EXIT_PRESS:
            goto OUT;
            break;

        default:
            break;
        }
    }

HANOI:
    hanoi(N, 0, 2, 1);

OUT:
#if 0
    cleardisplay();
    updatedisplay();
    BLIT(fd);
#endif

    funkyshutdown();
    DISPLAY(fd, 0);

    cleanup();

    exit(0);
}

int
rectangle(int x1, int y1, int x2, int y2, int outline, int fill)
{
    int i, j;
    int X1, X2, Y1, Y2;

    if (x2 < x1) {
        X1 = x2;
        X2 = x1;
    } else {
        X1 = x1;
        X2 = x2;
    }

    if (y2 < y1) {
        Y1 = y2;
        Y2 = y1;
    } else {
        Y1 = y1;
        Y2 = y2;
    }

    if ((X1 < 0) || (X2 < 0) || (X1 > EMPEG_XMAX) || (X2 > EMPEG_XMAX)) {
        fprintf(stderr, "invalid x coordinate(s): [(%d, %d), (%d, %d)]\n",
                        X1, Y1, X2, Y2);
        return -1;
    }

    if ((Y1 < 0) || (Y2 < 0) || (Y1 > EMPEG_YMAX) || (Y2 > EMPEG_YMAX)) {
        fprintf(stderr, "invalid y coordinate(s): [(%d, %d), (%d, %d)]\n",
                        X1, Y1, X2, Y2);
        return -1;
    }

    for (i = X1; i <= X2; i++) {
        display[i][Y1] = outline;
        display[i][Y2] = outline;
    }

    for (i = Y1; i <= Y2; i++) {
        display[X1][i] = outline;
        display[X2][i] = outline;
    }

    for (j = (Y1 + 1); j < Y2; j++) {
        for (i = (X1 + 1); i < X2; i++) {
            display[i][j] = fill;
        }
    }

    return 0;
}

/* Optimized rectangle() - not used at present */
#ifdef 0
int
rectangle_opt(int x1, int y1, int x2, int y2, int outline, int fill)
{
    int i, j, P;
    int X1, X2, Y1, Y2;
    int _X1, _X2;

    if (x2 < x1) {
        X1 = x2;
        X2 = x1;
    } else {
        X1 = x1;
        X2 = x2;
    }

    if (y2 < y1) {
        Y1 = y2;
        Y2 = y1;
    } else {
        Y1 = y1;
        Y2 = y2;
    }

    if ((X1 < 0) || (X2 < 0) || (X1 > EMPEG_XMAX) || (X2 > EMPEG_XMAX)) {
        fprintf(stderr, "invalid x coordinate(s): [(%d, %d), (%d, %d)]\n",
                        X1, Y1, X2, Y2);
        return -1;
    }

    if ((Y1 < 0) || (Y2 < 0) || (Y1 > EMPEG_YMAX) || (Y2 > EMPEG_YMAX)) {
        fprintf(stderr, "invalid y coordinate(s): [(%d, %d), (%d, %d)]\n",
                        X1, Y1, X2, Y2);
        return -1;
    }

    _X1 = X1;
    _X2 = X2;

    if (X1 % 2) { /* X1 odd */
        for (i = Y1; i <= Y2; i++) {
            P = dmap[((X1 - 1) / 2) + 64 * i];
            P = (P << 4) >> 4;
            dmap[((X1 - 1) / 2) + 64 * i] = P + (outline << 4);
        }
        X1 += 1;
        _X1 += 1;
    } else { /* X1 even */
        for (i = Y1; i <= Y2; i++) {
            dmap[(X1 / 2) + 64 * i] = outline + (fill << 4);
        }
        _X1 += 2;
    }

    if (!(X2 % 2)) { /* even */
        for (i = Y1; i <= Y2; i++) {
            P = dmap[(X2 / 2) + 64 * i];
            P = P >> 4;
            dmap[(X2 / 2) + 64 * i] = outline + (P << 4);
        }
        X2 -= 1;
        _X2 -= 1;
    } else {
        for (i = Y1; i <= Y2; i++) {
            dmap[((X2 - 1) / 2) + 64 * i] = fill + (outline << 4);
        }
        _X2 -= 2;
    }

    /* X1 is even, X2 is odd */

    for (i = X1 / 2; i < (X2 + 1) / 2; i++) {
        dmap[i + 64 * Y1] = outline + (outline << 4);
        dmap[i + 64 * Y2] = outline + (outline << 4);
    }

    for (i = (_X1 / 2); i < ((_X2 + 1) / 2); i++) {
        for (j = (Y1 + 1); j < Y2; j++) {
            dmap[i + 64 * j] = fill + (fill << 4);
        }
    }

    return 0;
}
#endif

void
movedisk_up(int diskid, int towerid)
{
    int D;
    int i, j, g;
    int xt1, xt2;
    int ax1, ay1, ax2, ay2;

    D = diskwidth(diskid);

    ax1 = towers[towerid].Txc - ((D + 1) / 2);
    ax2 = towers[towerid].Txc + ((D + 1) / 2);
    ay1 = Dy1(towers[towerid].ndisks - 1);
    ay2 = Dy2(towers[towerid].ndisks - 1);
    xt1 = towers[towerid].Txc - ((Tw - 1) / 2);
    xt2 = towers[towerid].Txc + ((Tw - 1) / 2);

    for (i = ay1; i > 0; i--) { /* off by 1 */
        if ((ay1 - i) > 3) {
            display[ax1][i + Dh - 1 + 1] = PX_BRI;
            display[ax1][i + Dh - 1 + 2] = PX_MED;
            display[ax1][i + Dh - 1 + 3] = PX_DIM;
            display[ax1][i + Dh - 1 + 4] = PX_OFF;
            display[ax2][i + Dh - 1 + 1] = PX_BRI;
            display[ax2][i + Dh - 1 + 2] = PX_MED;
            display[ax2][i + Dh - 1 + 3] = PX_DIM;
            display[ax2][i + Dh - 1 + 4] = PX_OFF;
        }
        rectangle(ax1, i, ax2, i + (Dh - 1), PX_OFF, PX_OFF);
        rectangle(ax1, i - 1, ax2, i + (Dh - 1) - 1, PX_MED, PX_DIM);
        g = i + (Dh - 1);
        if ((g >= (H_YMIN + Dh + 1)) && (g <= (H_YMAX - Tw - 1))) {
            for (j = xt1; j <= xt2; j++) {
                display[j][g] = PX_BRI;
            }
        }
        updatedisplay();
        BLIT(fd);
#ifdef GRAVITY
        usleep(movedelay + ffdelays[i - 1]);
#else
        usleep(movedelay);
#endif
    }

    rectangle(ax1, i, ax2, i + (Dh - 1), PX_OFF, PX_OFF);
    rectangle(ax1, 0, ax2, (Dh - 1), PX_MED, PX_DIM);

    display[ax1][i + Dh - 1 + 1] = PX_OFF;
    display[ax1][i + Dh - 1 + 2] = PX_OFF;
    display[ax1][i + Dh - 1 + 3] = PX_OFF;
    display[ax1][i + Dh - 1 + 4] = PX_OFF;
    display[ax2][i + Dh - 1 + 1] = PX_OFF;
    display[ax2][i + Dh - 1 + 2] = PX_OFF;
    display[ax2][i + Dh - 1 + 3] = PX_OFF;
    display[ax2][i + Dh - 1 + 4] = PX_OFF;

    updatedisplay();
    BLIT(fd);

    towers[towerid].ndisks--;
}

void
movedisk_lr(int diskid, int from, int to)
{
    int i;

#ifdef TRAILS
    int z = 0;
#endif

    int ax1, ay1, ax2, ay2;
    int bx1, bx2;
    int D = diskwidth(diskid);

    ax1 = towers[from].Txc - ((D + 1) / 2);
    ax2 = towers[from].Txc + ((D + 1) / 2);
    ay1 = 0;
    ay2 = Dh - 1;
    bx1 = towers[to].Txc - ((D + 1) / 2);
    bx2 = towers[to].Txc + ((D + 1) / 2);

    if (from < to) {
        for (i = ax1; i < bx1; i++) {

#ifdef TRAILS
            if (i > 3) {
                display[i - 1][ay1] = PX_BRI;
                display[i - 2][ay1] = PX_MED;
                display[i - 3][ay1] = PX_DIM;
                display[i - 4][ay1] = PX_OFF;
                display[i - 1][ay2] = PX_BRI;
                display[i - 2][ay2] = PX_MED;
                display[i - 3][ay2] = PX_DIM;
                display[i - 4][ay2] = PX_OFF;
            }
#endif
            rectangle(i, ay1, i + Dh + 1, ay2, PX_OFF, PX_OFF);
            rectangle(i + 1, ay1, i + D + 2, ay2, PX_MED, PX_DIM);
            updatedisplay();
            BLIT(fd);
            usleep(movedelay);
        }

#ifdef TRAILS
        display[i - 1][ay1] = PX_OFF;
        display[i - 2][ay1] = PX_OFF;
        display[i - 3][ay1] = PX_OFF;
        display[i - 4][ay1] = PX_OFF;
        display[i - 1][ay2] = PX_OFF;
        display[i - 2][ay2] = PX_OFF;
        display[i - 3][ay2] = PX_OFF;
        display[i - 4][ay2] = PX_OFF;
#endif

        rectangle(i + 1, ay1, i + D + 2, ay2, PX_OFF, PX_OFF);
    } else {
        for (i = ax1; i > bx1; i--) {

#ifdef TRAILS
            z = i + D + 1;
            display[z + 1][ay1] = PX_BRI;
            display[z + 2][ay1] = PX_MED;
            display[z + 3][ay1] = PX_DIM;
            display[z + 4][ay1] = PX_OFF;
            display[z + 1][ay2] = PX_BRI;
            display[z + 2][ay2] = PX_MED;
            display[z + 3][ay2] = PX_DIM;
            display[z + 4][ay2] = PX_OFF;
#endif
            rectangle(i, ay1, i + D + 1, ay2, PX_OFF, PX_OFF);
            rectangle(i - 1, ay1, i + D, ay2, PX_MED, PX_DIM);
            updatedisplay();
            BLIT(fd);
            usleep(movedelay);
        }

#ifdef TRAILS
        if (z > 0) { /* to shut warning off */
            display[z][ay1] = PX_OFF;
            display[z + 1][ay1] = PX_OFF;
            display[z + 2][ay1] = PX_OFF;
            display[z + 3][ay1] = PX_OFF;
            display[z + 4][ay1] = PX_OFF;
            display[z][ay2] = PX_OFF;
            display[z + 1][ay2] = PX_OFF;
            display[z + 2][ay2] = PX_OFF;
            display[z + 3][ay2] = PX_OFF;
            display[z + 4][ay2] = PX_OFF;
        }
#endif
        rectangle(i - 1, ay1, i + D, ay2, PX_OFF, PX_OFF);
    }

    rectangle(bx1, ay1, bx2, ay2, PX_MED, PX_DIM);
    updatedisplay();
    BLIT(fd);
}

void
movedisk_down(int diskid, int towerid)
{
    int D;
    int i, j, g, s = 0;
    int ax1, ay1, ax2, ay2;
    int by1, by2;
    int xt1, xt2;

    D = diskwidth(diskid);

    ax1 = towers[towerid].Txc - ((D + 1) / 2);
    ax2 = towers[towerid].Txc + ((D + 1) / 2);
    ay1 = 0;
    ay2 = Dh + 1;
    by1 = Dy1(towers[towerid].ndisks);
    by2 = Dy2(towers[towerid].ndisks);
    xt1 = towers[towerid].Txc - ((Tw - 1) / 2);
    xt2 = towers[towerid].Txc + ((Tw - 1) / 2);

    for (i = ay1; i < by1; i++, s++) { /* off by 1 */
        if (i > 3) {
            display[ax1][i - 1] = PX_BRI;
            display[ax1][i - 2] = PX_MED;
            display[ax1][i - 3] = PX_DIM;
            display[ax1][i - 4] = PX_OFF;
            display[ax2][i - 1] = PX_BRI;
            display[ax2][i - 2] = PX_MED;
            display[ax2][i - 3] = PX_DIM;
            display[ax2][i - 4] = PX_OFF;
        }
        rectangle(ax1, i, ax2, i + (Dh - 1), PX_OFF, PX_OFF);
        rectangle(ax1, i + 1, ax2, i + (Dh - 1) + 1, PX_MED, PX_DIM);
        updatedisplay();
        BLIT(fd);
        g = i;
        if ((g >= (H_YMIN + Dh + 1)) && (g <= (H_YMAX - Tw - 1))) {
            for (j = xt1; j <= xt2; j++) {
                display[j][g] = PX_BRI;
            }
        }
        updatedisplay();
        BLIT(fd);
#ifdef GRAVITY
        usleep(movedelay + ffdelays[s]);
#else
        usleep(movedelay);
#endif
    }

    display[ax1][i - 1] = PX_OFF;
    display[ax1][i - 2] = PX_OFF;
    display[ax1][i - 3] = PX_OFF;
    display[ax1][i - 4] = PX_OFF;
    display[ax2][i - 1] = PX_OFF;
    display[ax2][i - 2] = PX_OFF;
    display[ax2][i - 3] = PX_OFF;
    display[ax2][i - 4] = PX_OFF;
    updatedisplay();
    BLIT(fd);

    towers[towerid].ndisks++;
}

void
movedisk(int diskid, int from, int to)
{
    movedisk_up(diskid, from);
    movedisk_lr(diskid, from, to);
    movedisk_down(diskid, to);
    eh_moves++;
    showcounter(eh_moves);
#ifdef DEBUG
    getchar();
#endif
}

void
hanoi(int n, int from, int to, int using)
{
#define DISKID(n) (N - n)
    if (n > 0) {
        hanoi((n - 1), from, using, to);
        movedisk(DISKID(n), from, to);
        hanoi((n - 1), using, to, from);
    }
}

int
diskwidth(int diskid)
{
    int v;
    double d, r;

    d = Dw(diskid);
    r = fmod((double)d, (double)2);
    if (r == 0) { /* even */
        d = d + 1;
    }

    v = (int)d;

    if (v < (Tw + 2)) { /* ran out of tower space */
        return -1;
    }

    return v;
}

int
createtower(int towerid)
{
    int ret;
    int x1, y1, x2, y2;

    /* The base */

    x1 = towers[towerid].Txc - Tb2 - ((Tw - 1) / 2);
    x2 = towers[towerid].Txc + Tb2 + ((Tw - 1) / 2);
    y1 = Tby1;
    y2 = Tby2;

    ret = rectangle(x1, y1, x2, y2, PX_BRI, PX_BRI);

    if (ret != 0) {
        fprintf(stderr, "failed to draw base for tower %d\n", towerid);
        return -1;
    }

    /* The pole */

    x1 = towers[towerid].Txc - ((Tw - 1) / 2);
    x2 = towers[towerid].Txc + ((Tw - 1) / 2);
    y1 = H_YMIN + Dh + 1;
    y2 = H_YMAX - Tw - 1;

    ret = rectangle(x1, y1, x2, y2, PX_BRI, PX_BRI);

    if (ret != 0) {
        fprintf(stderr, "failed to draw pole for tower %d\n", towerid);
        return -1;
    }

    return 0;
}

void
cleardisplay(void)
{
    int i, j;

    for (i = 0; i < EMPEG_XMAX; i++)
        for (j = 0; j <= EMPEG_YMAX; j++)
            display[i][j] = 0;
}

void
updatedisplay(void)
{
    int i, j;

    pollbuttons();

    for (i = ((EMPEG_XMAX + 1)/2) - 1; i >= 0; i--)
        for (j = EMPEG_YMAX; j >= 0; j--)
            dmap[i + j * ((EMPEG_XMAX + 1)/2)] = display[2 * i][j] + 
                                                 (display[2 * i + 1][j] << 4);
}

void
popdisk(int towerid)
{
    int ret, diskid;
    int x1, y1, x2, y2;
    int xt1, xt2;
    int i, j;

    if (towers[towerid].ndisks == 0) {
        return;
    }

    diskid = towers[towerid].ndisks - 1;

    x1 = towers[towerid].Txc - ((diskwidth(diskid) + 1) / 2);
    x2 = towers[towerid].Txc + ((diskwidth(diskid) + 1) / 2);
    y1 = Dy1(diskid);
    y2 = Dy2(diskid);
    xt1 = towers[towerid].Txc - ((Tw - 1) / 2);
    xt2 = towers[towerid].Txc + ((Tw - 1) / 2);

    ret = rectangle(x1, y1, x2, y2, PX_OFF, PX_OFF);

    if (ret < 0) {
        fprintf(stderr, "failed to pop disk off tower %d\n", towerid);
        return;
    }

    for (i = y1; i <= y2; i++) {
        for (j = xt1; j <= xt2; j++) {
            display[j][i] = PX_BRI;
        }
    }

    towers[towerid].ndisks -= 1;

    updatedisplay();
    BLIT(fd);

    return;
}

int
createdisk(int diskid, int towerid)
{
    int ret;
    int x1, y1, x2, y2;

    x1 = towers[towerid].Txc - ((diskwidth(diskid) + 1) / 2);
    x2 = towers[towerid].Txc + ((diskwidth(diskid) + 1) / 2);
    y1 = Dy1(diskid);
    y2 = Dy2(diskid);

    ret = rectangle(x1, y1, x2, y2, PX_MED, PX_DIM);

    if (ret < 0) {
        fprintf(stderr, "failed to create disk %d on tower %d\n",
                diskid, towerid);
        return -1;
    }

    towers[towerid].ndisks += 1;

    updatedisplay();
    BLIT(fd);

    return 0;
}

void
pollbuttons(void)
{
    int ret, i, j;
    char buttonb[4];

    ret = read(ctrl, buttonb, 4);
    if (ret != 4) {
        return;
    }

    if ((buttonb[1] != 0) || (buttonb[2] != 0) || (buttonb[3] != 0)) {
        return;
    }

    switch (buttonb[0]) {
    case EXIT_PRESS:
        cleardisplay();
        for (i = ((EMPEG_XMAX + 1)/2) - 1; i >= 0; i--)
            for (j = EMPEG_YMAX; j >= 0; j--)
                dmap[i + j * ((EMPEG_XMAX + 1)/2)] = display[2 * i][j] + 
                             (display[2 * i + 1][j] << 4);
        BLIT(fd);
        DISPLAY(fd, 0);
        cleanup();
        exit(0);
        break;

    case VOLUME_UP:
        if (slowdown > 0) {
            slowdown--;
            movedelay = MOVEDELAY(slowdown);
            updatespeed();
        }
        break;

    case VOLUME_DOWN:
        if (slowdown < MAXSLOWDOWN) {
            slowdown++;
            movedelay = MOVEDELAY(slowdown);
            updatespeed();
        }
        break;
    }
}

void
cleanup(void)
{
    free_imatrix(display, 0, EMPEG_XMAX, 0, EMPEG_YMAX);

    if (munmap(dmap, 2048) == -1) {
        fprintf(stderr, "failed to munmap().\n");
    }

    close(fd);
    close(ctrl);
}

int
digit(int n, int x1, int y1)
{
    int i, j, a, b;

    if ((n > 9) || (n < 0)) {
        return -1;
    }

    a = b = 0;

    for (i = y1; i <= y1 + EHDFONT_HEIGHT - 1; i++, a++) {
        b = 0;
        for (j = x1; j <= x1 + EHDFONT_WIDTH - 1; j++, b++) {
            display[j][i] = digifontmap[n][a][b];
        }
    }

    return 0;
}

void
showcounter(int n)
{
    char S[4];
    int d0, d1, d2;

    if ((n < 0) || (n > 999)) {
        return;
    }

    sprintf(S, "%d", n);

    d2 = S[0] - '0';
    d1 = S[1] - '0';
    d0 = S[2] - '0';

    digit(d0, FDx1(1), FDy1(1));
    digit(d1, FDx1(2), FDy1(2));
    digit(d2, FDx1(3), FDy1(3));

    updatedisplay();
    BLIT(fd);
}

void
showlogo(void)
{
    int i, x1, y1;
    int u, v, a, b;

    for (i = 0; i < LOGOLEN; i++) {

        x1 = FLx1(i);
        y1 = FLy1(i);

        if (FLx2(i) > EMPEG_XMAX)
            break;

        a = b = 0;

        for (u = y1; u <= y1 + EHLFONT_HEIGHT - 1; u++, a++) {
            b = 0;
            for (v = x1; v <= x1 + EHLFONT_WIDTH - 1; v++, b++) {
                display[v][u] = charfontmap[i][a][b];
            }
        }
    }

    updatedisplay();
    BLIT(fd);
}

void
funkyshutdown(void)
{
    int i, j;

    movedelay = 100;

    for (i = 0; i <= EMPEG_YMAX; i += 2) {
        for (j = 0; j <= EMPEG_XMAX; j++) {
            display[j][i] = PX_OFF;
        }
        updatedisplay();
        BLIT(fd);
        usleep(movedelay);
    }

    for (i = EMPEG_YMAX; i >= 0; i -= 2) {
        for (j = 0; j <= EMPEG_XMAX; j++) {
            display[j][i] = PX_OFF;
        }
        updatedisplay();
        BLIT(fd);
        usleep(movedelay);
    }
}

void
showspeed(void)
{
    int i, x1, y1;
    int u, v, a, b;

    for (i = 0; i < SPEEDLEN; i++) {

        x1 = FSx1(i);
        y1 = FSy1(i);

        if (FSx2(i) > EMPEG_XMAX)
            break;

        a = b = 0;

        for (u = y1; u <= y1 + EHSFONT_HEIGHT - 1; u++, a++) {
            b = 0;
            for (v = x1; v <= x1 + EHSFONT_WIDTH - 1; v++, b++) {
                display[v][u] = speedfontmap[i][a][b];
            }
        }
    }

    updatespeed();
}

void
updatespeed(void)
{
    int z;
    int notches = MAXSLOWDOWN - slowdown;

    z = Fxmph + MAXSLOWDOWN;

    rectangle(Fxmph, Fymph, Fxmph + MAXSLOWDOWN, Fymph + Fhmph, PX_OFF, PX_OFF);
    rectangle(Fxmph - 1, Fymph - 1, Fxmph + MAXSLOWDOWN + 1, Fymph + Fhmph + 1,
              PX_DIM, PX_OFF);
    rectangle(Fxmph, Fymph, Fxmph + notches, Fymph + Fhmph, PX_BRI, PX_BRI);

    updatedisplay();
    BLIT(fd);
}

int
MOVEDELAY(int slowdown)
{
    int x = pow(DELAYPOWER, slowdown);
    x--;
    return x * DELAYMULTIPLIER;
}
/* FINIS */
