#ifndef _EMPHANOI_H
#define _EMPHANOI_H

/*
 * Copyright (C) 2002 Amit Singh <amit_s@acm.org>. All Rights Reserved.
 */

/*
 * User serviceable section is in config.h.
 */
#define MAXDISKS	8	/* absolutely maximum number of disks.     */
#define TRAILS		1	/* effect: disks leave trails behind them. */
#define GRAVITY		1	/* gravity emulation doesn't work well :-) */

#define MAXSLOWDOWN	5
#define DELAYMULTIPLIER	100
#define DELAYPOWER	6

/*
 * NOTE: The number of disks to be used for a Hanoi instance can be
 *       changed using the volume control knob on the Empeg. Clockwise
 *       rotation increases the number of disks and anti-clockwise
 *       decreases (bounded by 1 and MAXDISKS, both inclusive).
 *
 *       During animation, the volume control knob can be used to adjust
 *       animation speed.
 */

/*
 * Empeg constants.
 */

/* display dimensions */
#define EMPEG_XMAX	127
#define EMPEG_XMIN	0
#define EMPEG_YMAX	31
#define EMPEG_YMIN	0

/* devices */
#define DISPLAYDEV	"/dev/display"
#define CTRLDEV		"/dev/ir"

/* pixel intensity levels */
#define PX_OFF	0
#define PX_DIM	1
#define PX_MED	2
#define PX_BRI	3

/* button presses */
#define VOLUME_UP	0xa
#define VOLUME_DOWN	0xb
#define VOLUME_PRESS	0x8
#define VOLUME_RELEASE	0x9
#define RESET_PRESS	0x6
#define EXIT_PRESS	0x0

#define BLIT(fd)	ioctl((fd), _IO('d', 0))
#define DISPLAY(fd, x)	ioctl((fd), _IOW('d', 1, int), (x))

/*
 * Hanoi constants and formulas.
 */

/* animation bounding box */
#define H_XMAX 93
#define H_XMIN 0
#define H_YMAX 31
#define H_YMIN 0
#define H_XSIZE (H_XMAX - H_XMIN)
#define H_YSIZE (H_YMAX - H_YMIN)

/* standard disk thickness */
#define Dh	3

/* tower formulas */
#define Tb2	14
#define Tbw	4
#define Tby1	(H_YMAX - (Tbw - 1))
#define Tby2	(H_YMAX)
#define Ts	2
#define Tw	3

/* x-coordinates for tower centres */
#define T0xc	(H_XMIN + Tb2 + ((Dh + 1) / 2))
#define T1xc	(T0xc + 2 * Tb2 + ((Dh + 1 ) / 2) + Ts)
#define T2xc	(T1xc + 2 * Tb2 + ((Dh + 1 ) / 2) + Ts)

/* widths for bottom disk and tower base */
#define TBASE	(2 * Tb2 + Dh)
#define DBASE	(2 * Tb2)

/* width of the nth disk (the bottom-most disk has id 0) */
#define Dw(n)	(DBASE * \
                  ((pow((double)8, (double)n)) / (pow((double)10, (double)n))))

/*
 * For reference only: x-coordinates of the nth disk on the ith tower
 * 
 * Dx1i(n) = Tixc - ((diskwidth(n) + 1) / 2)
 * Dx2i(n) = Tixc + ((diskwidth(n) + 1) / 2)
 */

/* y-coordinates of the nth disk */
#define Dy1(n)	(Tby1 - (((n) + 1) * Dh))
#define Dy2(n)	(Tby1 - ((n) * Dh) - 1)

/*
 * Data structures and prototypes.
 */

typedef struct tower_s {
    int id;
    int ndisks;
    int Txc;
    int topdisk;
} tower_t;

int MOVEDELAY(int);
void cleanup(void);
void showlogo(void);
int diskwidth(int id);
void showspeed(void);
void pollbuttons(void);
void updatespeed(void);
void showcounter(int n);
int createtower(int id);
void cleardisplay(void);
void updatedisplay(void);
void funkyshutdown(void);
void popdisk(int towerid);
int digit(int n, int x1, int y1);
int createdisk(int diskid, int towerid);
void movedisk_up(int diskid, int towerid);
void movedisk_down(int diskid, int towerid);
void movedisk(int diskid, int from, int to);
void movedisk_lr(int diskid, int from, int to);
void hanoi(int ndisks, int from, int to, int using);
int rectangle(int x1, int y1, int x2, int y2, int outline, int fill);

#endif /* _EMPHANOI_H */
